Package: familiar
Title: End-to-End Automated Machine Learning and Model Evaluation
Version: 2.0.2
Authors@R: c(
    person("Alex", "Zwanenburg",
           email = "alexander.zwanenburg@nct-dresden.de",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-0342-9545")),
    person("Steffen", "Löck", role="aut"),
    person("German Cancer Research Center (DKFZ)", role="cph"),
    person("Technische Universität Dresden", role="cph"))
Description: Single unified interface for end-to-end modelling of regression, 
    categorical and time-to-event (survival) outcomes. Models created using
    familiar are self-containing, and their use does not require additional
    information such as baseline survival, feature clustering, or feature
    transformation and normalisation parameters. Model performance,
    calibration, risk group stratification, (permutation) variable importance,
    individual conditional expectation, partial dependence, and more, are
    assessed automatically as part of the evaluation process and exported in
    tabular format and plotted, and may also be computed manually using export
    and plot functions. Where possible, metrics and values obtained during the
    evaluation process come with confidence intervals.
URL: https://github.com/oncoray/familiar
BugReports: https://github.com/oncoray/familiar/issues
Depends: R (>= 4.0.0)
License: EUPL
Encoding: UTF-8
VignetteBuilder: knitr
Imports: data.table, methods, rlang (>= 1.0.0), rstream, survival
Suggests: BART, callr (>= 3.4.3), cluster, CORElearn, coro,
        dynamicTreeCut, e1071 (>= 1.7.5), fastcluster, fastglm, ggplot2
        (>= 4.0.0), glmnet, gtable, harmonicmeanp, isotree (>= 0.6.0),
        knitr, labeling, laGP, maxstat, microbenchmark, nnet,
        paletteer, power.transform, praznik, proxy, randomForestSRC,
        ranger, rmarkdown, scales, testthat (>= 3.0.0), xml2, xgboost
        (>= 3.0.0)
Collate: 'FamiliarS4Classes.R' 'FamiliarS4Generics.R'
        'BatchNormalisation.R' 'BootstrapConfidenceInterval.R'
        'CheckArguments.R' 'CheckHyperparameters.R' 'CheckPackages.R'
        'ClassBalance.R' 'ClusteringMethod.R' 'Clustering.R'
        'ClusterRepresentation.R' 'Normalisation.R'
        'CombatNormalisation.R' 'LearnerS4Naive.R' 'DataObject.R'
        'DataParameterChecks.R' 'DataPreProcessing.R'
        'DataProcessing.R' 'DataServerBackend.R' 'ErrorMessages.R'
        'ExperimentData.R' 'ExperimentSetup.R' 'Familiar.R'
        'FamiliarCollection.R' 'FamiliarCollectionExport.R'
        'FamiliarData.R' 'FamiliarDataComputation.R'
        'FamiliarDataComputationPredictionData.R' 'PredictionTable.R'
        'FamiliarDataComputationAUCCurves.R'
        'FamiliarDataComputationCalibrationData.R'
        'FamiliarDataComputationCalibrationInfo.R'
        'FamiliarDataComputationConfusionMatrix.R'
        'FamiliarDataComputationDecisionCurveAnalysis.R'
        'FamiliarDataComputationFeatureExpression.R'
        'FamiliarDataComputationFeatureSimilarity.R'
        'FamiliarDataComputationHyperparameters.R'
        'FamiliarDataComputationICE.R'
        'FamiliarDataComputationModelPerformance.R'
        'FamiliarDataComputationPermutationVimp.R'
        'FamiliarDataComputationRiskStratificationData.R'
        'FamiliarDataComputationRiskStratificationInfo.R'
        'FamiliarDataComputationSHAP.R'
        'FamiliarDataComputationSampleSimilarity.R'
        'FamiliarDataComputationUnivariateAnalysis.R'
        'FamiliarDataComputationUtilities.R'
        'FamiliarDataComputationVimp.R' 'FamiliarDataElement.R'
        'FamiliarEnsemble.R' 'FamiliarHyperparameterLearner.R'
        'FamiliarModel.R' 'FamiliarNoveltyDetector.R'
        'FamiliarObjectConversion.R' 'Transformation.R'
        'FamiliarObjectUpdate.R' 'FamiliarSharedS4Methods.R'
        'FamiliarVimpMethod.R' 'FeatureInfo.R'
        'FeatureInfoParameters.R' 'FunctionWrapperUtilities.R'
        'HyperparameterOptimisation.R'
        'HyperparameterOptimisationMetaLearners.R'
        'HyperparameterOptimisationUtilities.R'
        'HyperparameterS4BayesianAdditiveRegressionTrees.R'
        'HyperparameterS4GaussianProcess.R'
        'HyperparameterS4RandomSearch.R' 'HyperparameterS4Ranger.R'
        'Imputation.R' 'Iterations.R' 'LearnerMain.R'
        'LearnerRecalibration.R' 'LearnerRiskStratification.R'
        'LearnerS4Cox.R' 'LearnerS4GLM.R' 'LearnerS4GLMnet.R'
        'LearnerS4KNN.R' 'LearnerS4MBoost.R' 'LearnerS4NaiveBayes.R'
        'LearnerS4RFSRC.R' 'LearnerS4Ranger.R' 'LearnerS4SVM.R'
        'LearnerS4SurvivalRegression.R' 'LearnerS4XGBoost.R'
        'LearnerSurvivalProbability.R' 'Logger.R' 'MetricS4.R'
        'MetricS4AUC.R' 'MetricS4Brier.R' 'MetricS4ConcordanceIndex.R'
        'MetricS4ConfusionMatrixMetrics.R' 'MetricS4Regression.R'
        'NoveltyDetectorMain.R' 'NoveltyDetectorS4IsolationTree.R'
        'NoveltyDetectorS4NoneNoveltyDetector.R' 'OutcomeInfo.R'
        'PairwiseSimilarity.R' 'ParallelFunctions.R' 'ParseData.R'
        'ParseSettings.R' 'PlotAUCcurves.R' 'PlotAll.R'
        'PlotCalibration.R' 'PlotColours.R' 'PlotConfusionMatrix.R'
        'PlotDecisionCurves.R' 'PlotFamiliarPlot.R'
        'PlotFeatureRanking.R' 'PlotFeatureSimilarity.R' 'PlotGTable.R'
        'PlotICE.R' 'PlotInputArguments.R' 'PlotKaplanMeier.R'
        'PlotModelPerformance.R' 'PlotPermutationVariableImportance.R'
        'PlotSampleClustering.R' 'PlotShapDependence.R'
        'PlotShapForce.R' 'PlotShapSummary.R' 'PlotShapWaterfall.R'
        'PlotUnivariateImportance.R' 'PlotUtilities.R'
        'PredictS4Methods.R' 'ProcessTimeUtilities.R' 'Random.R'
        'RankBordaAggregation.R' 'RankMain.R' 'RankSimpleAggregation.R'
        'RankStabilityAggregation.R' 'SocketServer.R'
        'StringUtilities.R' 'TaskEvaluate.R' 'TaskFeatureInfo.R'
        'TaskLearn.R' 'TaskLearnerHyperparameters.R' 'TaskMain.R'
        'TaskNoveltyDetector.R' 'TaskNoveltyDetectorHyperparameters.R'
        'TaskVimp.R' 'TaskVimpHyperparameters.R' 'TestDataCreators.R'
        'TestFeatureInfo.R' 'TestFunctions.R' 'TestTrain.R'
        'TestTrainNovelty.R' 'TestVimp.R' 'TrimUtilities.R'
        'Utilities.R' 'UtilitiesS4.R' 'VimpMain.R'
        'VimpS4Concordance.R' 'VimpS4CoreLearn.R' 'VimpS4Correlation.R'
        'VimpS4MutualInformation.R' 'VimpS4OtherMethods.R'
        'VimpS4Regression.R' 'VimpTable.R' 'aaa.R'
Config/testthat/parallel: true
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-01 15:43:01 UTC; alexz
Author: Alex Zwanenburg [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0342-9545>),
  Steffen Löck [aut],
  German Cancer Research Center (DKFZ) [cph],
  Technische Universität Dresden [cph]
Maintainer: Alex Zwanenburg <alexander.zwanenburg@nct-dresden.de>
Repository: CRAN
Date/Publication: 2026-06-01 17:30:02 UTC
Built: R 4.6.0; ; 2026-06-01 18:07:31 UTC; unix
