# Generated by roxygen2: do not edit by hand

export(as_prediction_table)
export(coef)
export(get_xml_config)
export(plot_feature_selection_occurrence)
export(plot_feature_selection_variable_importance)
export(plot_model_signature_occurrence)
export(plot_model_signature_variable_importance)
export(precompute_data_assignment)
export(precompute_feature_info)
export(precompute_vimp)
export(summon_familiar)
export(theme_familiar)
export(train_familiar)
export(vcov)
export(waiver)
exportClasses(experimentData)
exportClasses(familiarCollection)
exportClasses(familiarData)
exportClasses(familiarDataElement)
exportClasses(familiarEnsemble)
exportClasses(familiarHyperparameterLearner)
exportClasses(familiarMetric)
exportClasses(familiarModel)
exportClasses(familiarNoveltyDetector)
exportClasses(familiarVimpMethod)
exportClasses(featureInfo)
exportClasses(featureInfoParameters)
exportClasses(outcomeInfo)
exportClasses(vimpTable)
exportMethods(aggregate_vimp_table)
exportMethods(as_data_object)
exportMethods(as_familiar_collection)
exportMethods(as_familiar_data)
exportMethods(as_familiar_ensemble)
exportMethods(export_all)
exportMethods(export_auc_data)
exportMethods(export_calibration_data)
exportMethods(export_calibration_info)
exportMethods(export_confusion_matrix_data)
exportMethods(export_decision_curve_analysis_data)
exportMethods(export_feature_expressions)
exportMethods(export_feature_similarity)
exportMethods(export_fs_vimp)
exportMethods(export_hyperparameters)
exportMethods(export_ice_data)
exportMethods(export_model_performance)
exportMethods(export_model_vimp)
exportMethods(export_partial_dependence_data)
exportMethods(export_permutation_vimp)
exportMethods(export_prediction_data)
exportMethods(export_risk_stratification_data)
exportMethods(export_risk_stratification_info)
exportMethods(export_sample_similarity)
exportMethods(export_univariate_analysis_data)
exportMethods(get_class_names)
exportMethods(get_data_set_names)
exportMethods(get_feature_names)
exportMethods(get_learner_names)
exportMethods(get_risk_group_names)
exportMethods(get_vimp_method_names)
exportMethods(get_vimp_table)
exportMethods(plot_auc_precision_recall_curve)
exportMethods(plot_auc_roc_curve)
exportMethods(plot_calibration_data)
exportMethods(plot_confusion_matrix)
exportMethods(plot_decision_curve)
exportMethods(plot_feature_similarity)
exportMethods(plot_ice)
exportMethods(plot_kaplan_meier)
exportMethods(plot_model_performance)
exportMethods(plot_pd)
exportMethods(plot_permutation_variable_importance)
exportMethods(plot_sample_clustering)
exportMethods(plot_shap_dependence)
exportMethods(plot_shap_force)
exportMethods(plot_shap_summary)
exportMethods(plot_shap_waterfall)
exportMethods(plot_univariate_importance)
exportMethods(plot_variable_importance)
exportMethods(predict)
exportMethods(set_class_names)
exportMethods(set_data_set_names)
exportMethods(set_feature_names)
exportMethods(set_learner_names)
exportMethods(set_risk_group_names)
exportMethods(set_vimp_method_names)
exportMethods(summary)
exportMethods(update_model_dir_path)
exportMethods(update_object)
import(data.table)
import(methods)
importClassesFrom(rstream,rstream.runif)
importFrom(rlang,"%<~%")
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,ensym)
importFrom(rlang,ensyms)
importFrom(rlang,parse_expr)
importFrom(rlang,parse_exprs)
importFrom(rlang,quo)
importFrom(rlang,quos)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(stats,coef)
importFrom(stats,predict)
importFrom(stats,vcov)
importFrom(survival,Surv)
importFrom(survival,coxph)
importFrom(survival,survreg)
importFrom(utils,head)
importFrom(utils,tail)
