# Familiar

A list of tools to help create D&D applications.

## Installation

Run the following to install:

```python

pip install familiar
```

## Usage

```python
from familiar import dice_roll

# dice_roll takes a number of dice to roll, number of sides on dice, and a modifier and returns a random result.

roll = dice_roll(2,6, 1)
print(roll)
```
