% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_data_set_names,familiarCollection-method}
\alias{get_data_set_names,familiarCollection-method}
\alias{get_data_set_names}
\title{Get current name of datasets}
\usage{
\S4method{get_data_set_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of dataset name labels.
}
\description{
Datasets in familiarCollection objects can have custom names for
export and plotting. This function retrieves the currently assigned names.
}
\details{
Labels convert internal naming of data sets to the requested label
at export or when plotting. Labels can be changed using the
\code{set_data_set_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_data_set_names}} for updating the name of datasets and their ordering.
}
}
