% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_class_names,familiarCollection-method}
\alias{get_class_names,familiarCollection-method}
\alias{get_class_names}
\title{Get outcome class labels}
\usage{
\S4method{get_class_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of class labels.
}
\description{
Outcome classes in familiarCollection objects can have custom
names for export and plotting. This function retrieves the currently
assigned names.
}
\details{
Labels convert internal class names to the requested label at export
or when plotting. Labels can be changed using the \code{set_class_names}
method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_class_names}} for updating the name and ordering of classes.
}
}
