% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{delayedDataObject-class}
\alias{delayedDataObject-class}
\title{Data object with delayed loading}
\description{
The delayed loading object provides an interface to the backend data. This
data object is typically used within the evaluation pipeline to load data
when needed.
}
\section{Slots}{

\describe{
\item{\code{data}}{NULL or data table containing the data. If present (not \code{NULL}),
data is considered loaded. This should not happen -- load_data_object auto-
matically creates a dataObject from the delayedDataObject.}

\item{\code{preprocessing_level}}{character indicating the level of pre-processing
already conducted. \code{"none"} by default.}

\item{\code{outcome_type}}{character, determines the outcome type.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters. Optional.}

\item{\code{data_column_info}}{Object containing column information.}

\item{\code{data_id}}{integer. Defines the data_id of the dataset that should be
loaded.}

\item{\code{run_id}}{integer. Defines the run_id of the dataset that should be load.
Together with data_id, run_id and validation allows for looking up the
sample set. If run_id is left unset (NA_integer_), this will force the
run_id to be set using the model, vimp_method or ensemble object. This is
used during the evaluation process to load data specifically related to
training, internal validation and external validation. The run-tables
(which contain information about data partitioning) associated with these
objects are used to look-up the run_id based on the data_id (that is always
explicitly set). The perform_task method for familiarTaskEvaluate uses this
aspect explicitly.}

\item{\code{validation}}{logical. This determines which internal data set will be
loaded. If TRUE, the validation data will be loaded, whereas FALSE loads
the development data.}

\item{\code{aggregate_on_load}}{logical. Determines whether data is aggregated after
loading.}

\item{\code{sample_set_on_load}}{NULL or vector of sample identifiers to be loaded.
Overrides any \code{sample_seed} that may have been provided.}
}}

