% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{is.encapsulated_path}
\alias{is.encapsulated_path}
\title{Internal test for encapsulated_path}
\usage{
is.encapsulated_path(x)
}
\arguments{
\item{x}{Object to be tested.}
}
\value{
\code{TRUE} for objects that are \code{encapsulated_path}, \code{FALSE} otherwise.
}
\description{
This function tests if the object is an \code{encapsulated_path} object.
}
\keyword{internal}
