% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Familiar.R
\docType{package}
\name{familiar}
\alias{familiar-package}
\alias{familiar}
\title{familiar: Fully Automated Machine Learning with Interpretable Analysis of Results}
\description{
End-to-end, automated machine learning package for creating
trustworthy and interpretable machine learning models. Familiar supports
modelling of regression, categorical and time-to-event (survival) outcomes.
Models created using familiar are self-containing, and their use does not
require additional information such as baseline survival, feature
clustering, or feature transformation and normalisation parameters. In
addition, an novelty or out-of-distribution detector is trained
simultaneously and contained with every model. Model performance,
calibration, risk group stratification, (permutation) variable importance,
individual conditional expectation, partial dependence, and more, are
assessed automatically as part of the evaluation process and exported in
tabular format and plotted, and may also be computed manually using export
and plot functions. Where possible, metrics and values obtained during the
evaluation process come with confidence intervals.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/oncoray/familiar}
  \item Report bugs at \url{https://github.com/oncoray/familiar/issues}
}

}
\author{
\strong{Maintainer}: Alex Zwanenburg \email{alexander.zwanenburg@nct-dresden.de} (\href{https://orcid.org/0000-0002-0342-9545}{ORCID})

Authors:
\itemize{
  \item Steffen Löck
}

Other contributors:
\itemize{
  \item German Cancer Research Center (DKFZ) [copyright holder]
  \item Technische Universität Dresden [copyright holder]
}

}
