% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarObjectUpdate.R
\name{update_object}
\alias{update_object}
\alias{update_object,familiarModel-method}
\alias{update_object,familiarEnsemble-method}
\alias{update_object,familiarData-method}
\alias{update_object,familiarCollection-method}
\alias{update_object,vimpTable-method}
\alias{update_object,familiarNoveltyDetector-method}
\alias{update_object,featureInfo-method}
\alias{update_object,featureInfoParametersTransformationPowerTransform-method}
\alias{update_object,experimentData-method}
\alias{update_object,list-method}
\alias{update_object,ANY-method}
\title{Update familiar S4 objects to the most recent version.}
\usage{
update_object(object, ...)

\S4method{update_object}{familiarModel}(object, ...)

\S4method{update_object}{familiarEnsemble}(object, ...)

\S4method{update_object}{familiarData}(object, ...)

\S4method{update_object}{familiarCollection}(object, ...)

\S4method{update_object}{vimpTable}(object, ...)

\S4method{update_object}{familiarNoveltyDetector}(object, ...)

\S4method{update_object}{featureInfo}(object, ...)

\S4method{update_object}{featureInfoParametersTransformationPowerTransform}(object, ...)

\S4method{update_object}{experimentData}(object, ...)

\S4method{update_object}{list}(object, ...)

\S4method{update_object}{ANY}(object, ...)
}
\arguments{
\item{object}{A \code{familiarModel}, a \code{familiarEnsemble}, a \code{familiarData} or
\code{familiarCollection} object.}

\item{...}{Unused arguments.}
}
\value{
An up-to-date version of the respective S4 object.
}
\description{
Provides backward compatibility for familiar objects exported to
a file. This mitigates compatibility issues when working with files that
become outdated as new versions of familiar are released, e.g. because
slots have been removed.

Major version releases (e.g., versions 1.0.0, 2.0.0) introduce breaking
changes which can result in objects that cannot be updated.
}
