% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarModel.R
\name{summary}
\alias{summary}
\alias{summary,familiarModel-method}
\title{Model summaries}
\usage{
summary(object, ...)

\S4method{summary}{familiarModel}(object, ...)
}
\arguments{
\item{object}{a familiarModel object}

\item{...}{additional arguments passed to \code{summary} methods for the underlying
model, when available.}
}
\value{
Depends on underlying model. See the documentation for the particular
models.
}
\description{
\code{summary} produces model summaries.
}
\details{
This method extends the \code{summary} S3 method. For some models
\code{summary} requires information that is trimmed from the model. In this case
a copy of summary data is stored with the model, and returned.
}
