% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationHyperparameters.R
\name{extract_hyperparameters}
\alias{extract_hyperparameters}
\title{Internal function to extract hyperparameters from models.}
\usage{
extract_hyperparameters(object, message_indent = 0L, verbose = FALSE, ...)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A list of familiarDataElements with hyperparameters.
}
\description{
Collects hyperparameters from models in a \code{familiarEnsemble}.
}
\keyword{internal}
