% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarNoveltyDetector.R
\name{set_object_name,familiarNoveltyDetector-method}
\alias{set_object_name,familiarNoveltyDetector-method}
\title{Set the name of a \code{familiarNoveltyDetector} object.}
\usage{
\S4method{set_object_name}{familiarNoveltyDetector}(x, new = NULL)
}
\arguments{
\item{x}{A \code{familiarNoveltyDetector} object.}
}
\value{
A \code{familiarNoveltyDetector} object with a generated or a provided name.
}
\description{
Set the \code{name} slot using the object name.
}
\keyword{internal}
