% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{is.waive}
\alias{is.waive}
\title{Internal test to see if an object is a waiver}
\usage{
is.waive(x)
}
\arguments{
\item{x}{Object to be tested.}
}
\value{
\code{TRUE} for objects that are waivers, \code{FALSE} otherwise.
}
\description{
This function tests if the object was created by the \code{waiver} function. This
function is functionally identical to \code{ggplot2:::is.waive()}.
}
\keyword{internal}
