% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseData.R
\name{.parse_integer_features}
\alias{.parse_integer_features}
\title{Internal function for converting integer features}
\usage{
.parse_integer_features(data, outcome_type)
}
\arguments{
\item{data}{data.table with feature data}

\item{outcome_type}{character, indicating the type of outcome}
}
\value{
data.table with integer features converted to double.
}
\description{
Internal function for converting integer features
}
\details{
This function parses columns containing integer feature data to
features to double. This prevents, e.g., errors when the result of an
operation on the feature data yields a non-integer (i.e. floating point)
result.
}
\keyword{internal}
