% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{dataObject-class}
\alias{dataObject-class}
\title{Data object}
\description{
The dataObject class is used to resolve the issue of keeping track of
pre-processing status and data loading inside complex workflows, e.g. nested
predict functions inside a calibration function.
}
\section{Slots}{

\describe{
\item{\code{data}}{NULL or data table containing the data. This is the data which
will be read and used.}

\item{\code{preprocessing_level}}{character indicating the level of pre-processing
already conducted.}

\item{\code{outcome_type}}{character, determines the outcome type.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters. Optional.}

\item{\code{data_column_info}}{Object containing column information.}

\item{\code{data_id}}{Data identifier for dataset. Set using internal routines if the
\code{dataObject} was created from a \code{delayedDataObject}}

\item{\code{run_id}}{Run identifier for dataset. Set using internal routines if the
\code{dataObject} was created from a \code{delayedDataObject}}

\item{\code{validation}}{Identifies if validation or development samples were loaded.
Set using internal routines if the \code{dataObject} was created from a
\code{delayedDataObject}.}

\item{\code{sample_seed}}{Seed used for creating a bootstrap of the data.}
}}

