% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.check_feature_availability}
\alias{.check_feature_availability}
\title{Internal function to check whether feature columns are found in the data}
\usage{
.check_feature_availability(data, feature)
}
\arguments{
\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{feature}{Character string(s) indicating one or more features.}
}
\description{
This function checks whether feature columns can be found in the data set.
It will raise an error if any feature columns are missing from the data set.
}
\keyword{internal}
