% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarModel.R
\name{vcov}
\alias{vcov}
\alias{vcov,familiarModel-method}
\title{Calculate variance-covariance matrix for a model}
\usage{
vcov(object, ...)

\S4method{vcov}{familiarModel}(object, ...)
}
\arguments{
\item{object}{a familiarModel object}

\item{...}{additional arguments passed to \code{vcov} methods for the underlying
model, when available.}
}
\value{
Variance-covariance matrix of the model in the familiarModel object,
if any.
}
\description{
Calculate variance-covariance matrix for a model
}
\details{
This method extends the \code{vcov} S3 method. For some models \code{vcov}
requires information that is trimmed from the model. In this case a copy of
the variance-covariance matrix is stored with the model, and returned.
}
