% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{waiver}
\alias{waiver}
\title{Create a waiver object}
\usage{
waiver()
}
\value{
waiver object
}
\description{
This function is functionally identical to \code{ggplot2::waiver()} function and
creates a waiver object. A waiver object is an otherwise empty object that
serves the same purpose as \code{NULL}, i.e. as placeholder for a default value.
Because \code{NULL} can sometimes be a valid input argument, it can therefore not
be used to switch to an internal default value.
}
