% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarNoveltyDetector-class}
\alias{familiarNoveltyDetector-class}
\title{Novelty detector.}
\description{
A familiarNoveltyDetector object is a self-contained model that can be
applied to generate out-of-distribution predictions for instances in a
dataset.
}
\details{
Note that these objects do not contain any data concerning outcome,
as this not relevant for (prospective) out-of-distribution detection.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the familiarNoveltyDetector object.}

\item{\code{model}}{The actual novelty detector trained using a specific algorithm,
e.g. a isolation forest from the \code{isotree} package.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters.}

\item{\code{data_column_info}}{Data information object containing information
regarding identifier column names.}

\item{\code{hyperparameters}}{Set of hyperparameters used to train the detector.}

\item{\code{hyperparameter_data}}{Information generated during hyperparameter
optimisation. Currently not used.}

\item{\code{calibration_model}}{Model used to convert raw output to statistical
probability of being out-of-distribution. Currently not used.}

\item{\code{learner}}{Learning algorithm used to create the novelty detector.}

\item{\code{vimp_method}}{Method used to determine variable importance for the
novelty detector.}

\item{\code{vimp_table}}{Variable importance table or list of variable importance
tables for the model.}

\item{\code{vimp_aggregation_method}}{Method used for aggregating variable importance
tables if more than one is present.)}

\item{\code{vimp_rank_threshold}}{Threshold used for some variable importance
aggregation methods.}

\item{\code{required_features}}{The set of features required for complete
reproduction, i.e. with imputation.}

\item{\code{model_features}}{The set of features that is used to train the detector.}

\item{\code{data_id}}{Internal identifier for the dataset used to train the detector.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
used to train the detector.}

\item{\code{run_table}}{Run table for the data used to train the detector. Used
internally.}

\item{\code{is_trimmed}}{Flag that indicates whether the detector, stored in the
\code{model} slot, has been trimmed.}

\item{\code{trimmed_function}}{List of functions whose output has been captured prior
to trimming the model.}

\item{\code{messages}}{List of warning and error messages generated during training.}

\item{\code{project_id}}{Identifier of the project that generated the
familiarNoveltyDetector object.}

\item{\code{familiar_version}}{Version of the familiar package.}

\item{\code{package}}{Name of package(s) required to executed the detector itself,
e.g. \code{isotree}.}

\item{\code{package_version}}{Version of the packages mentioned in the \code{package}
attribute.}
}}

