% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_vimp_method_names,familiarCollection-method}
\alias{get_vimp_method_names,familiarCollection-method}
\alias{get_vimp_method_names}
\title{Get current variable importance method name labels}
\usage{
\S4method{get_vimp_method_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of variable importance method name labels.
}
\description{
Variable importance methods in familiarCollection objects can
have custom names for export and plotting. This function retrieves the
currently assigned names.
}
\details{
Labels convert internal naming of variable importance methods to the
requested label at export or when plotting. Labels can be changed using the
\code{set_vimp_method_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_vimp_method_names}} for updating the name of variable importance methods and their ordering.
}
}
