% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarMetric-class}
\alias{familiarMetric-class}
\title{Model performance metric.}
\description{
Superclass for model performance objects.
}
\section{Slots}{

\describe{
\item{\code{metric}}{Performance metric.}

\item{\code{outcome_type}}{Type of outcome being predicted.}

\item{\code{name}}{Name of the performance metric.}

\item{\code{value_range}}{Range of the performance metric. Can be half-open.}

\item{\code{baseline_value}}{Value of the metric for trivial models, e.g. models that
always predict the median value, the majority class, or the mean hazard,
etc.}

\item{\code{higher_better}}{States whether higher metric values correspond to better
predictive model performance (e.g. accuracy) or not (e.g. root mean squared
error).}
}}

