% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Iterations.R
\name{.load_iterations}
\alias{.load_iterations}
\title{Internal function for loading iteration data from the file system}
\usage{
.load_iterations(file_dir, iteration_file = NULL)
}
\arguments{
\item{file_dir}{Path to directory where iteration files are stored.}
}
\value{
List containing:
\itemize{
\item \code{iteration_file_exists}: An indicator whether an iteration file was found.
\item \code{iteration_list}: The list of iterations (if available).
\item \code{project_id}: The unique project identifier (if available).
}
}
\description{
Loads iterations generated by \code{.create_iterations} that were created in a
previous session. If these are not available, this is indicated by setting a
return flag.
}
\keyword{internal}
