% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerMain.R
\name{.get_default_sign_size}
\alias{.get_default_sign_size}
\title{Internal function for obtaining a default signature size parameter}
\usage{
.get_default_sign_size(data, restrict_samples = FALSE)
}
\arguments{
\item{data}{dataObject class object which contains the data on which the
preset parameters are determined.}

\item{restrict_samples}{Logical indicating whether the signature size should
be limited by the number of samples in addition to the number of available
features. This may help convergence of OLS-based methods.}
}
\value{
List containing the preset values for the signature size parameter.
}
\description{
Internal function for obtaining a default signature size parameter
}
\keyword{internal}
