% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotUtilities.R
\name{theme_familiar}
\alias{theme_familiar}
\title{Familiar ggplot2 theme}
\usage{
theme_familiar(
  base_size = 10,
  base_family = "",
  base_line_size = 0.5,
  base_rect_size = 0.5
)
}
\arguments{
\item{base_size}{Base font size in points. Size of other plot text elements
is based off this.}

\item{base_family}{Font family used for text elements.}

\item{base_line_size}{Base size for line elements, in points.}

\item{base_rect_size}{Base size for rectangular elements, in points.}
}
\value{
A complete plotting theme.
}
\description{
This is the default theme used for plots created by familiar. The theme uses
\code{ggplot2::theme_light} as the base template.
}
