% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarObjectConversion.R
\name{as_familiar_ensemble}
\alias{as_familiar_ensemble}
\alias{as_familiar_ensemble,familiarEnsemble-method}
\alias{as_familiar_ensemble,familiarModel-method}
\alias{as_familiar_ensemble,familiarNoveltyDetector-method}
\alias{as_familiar_ensemble,list-method}
\alias{as_familiar_ensemble,character-method}
\alias{as_familiar_ensemble,ANY-method}
\title{Conversion to familiarEnsemble object.}
\usage{
as_familiar_ensemble(object, ...)

\S4method{as_familiar_ensemble}{familiarEnsemble}(object, ...)

\S4method{as_familiar_ensemble}{familiarModel}(object, ...)

\S4method{as_familiar_ensemble}{familiarNoveltyDetector}(object, ...)

\S4method{as_familiar_ensemble}{list}(object, ...)

\S4method{as_familiar_ensemble}{character}(object, ...)

\S4method{as_familiar_ensemble}{ANY}(object, ...)
}
\arguments{
\item{object}{A \code{familiarEnsemble} object, or one or more \code{familiarModel}
objects that will be internally converted to a \code{familiarEnsemble} object.
Paths to such objects can also be provided.}

\item{...}{Unused arguments.}
}
\value{
A \code{familiarEnsemble} object.
}
\description{
Creates \code{familiarEnsemble} a object from \code{familiarModel}
objects.
}
