% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarVimpMethod-class}
\alias{familiarVimpMethod-class}
\title{Variable importance method object.}
\description{
The familiarVimpMethod class is the parent class for all variable importance
methods in familiar.
}
\section{Slots}{

\describe{
\item{\code{outcome_type}}{Outcome type of the data to be evaluated using the object.}

\item{\code{hyperparameters}}{Set of hyperparameters for the variable importance
method.}

\item{\code{vimp_method}}{The character string indicating the variable importance
method.}

\item{\code{vimp_aggregation_method}}{Method used for aggregating variable importance
tables if more than one is present.)}

\item{\code{vimp_rank_threshold}}{Threshold used for some variable importance
aggregation methods.}

\item{\code{multivariate}}{Flags whether the variable importance method is
multivariate vs. univariate.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters.}

\item{\code{required_features}}{The set of features to be assessed by the variable
importance method.}

\item{\code{package}}{Name of the package(s) required to execute the variable
importance method.}

\item{\code{run_table}}{Run table for the data to be assessed by the variable
importance method. Used internally.}

\item{\code{project_id}}{Identifier of the project that generated the
familiarVimpMethod object.}

\item{\code{familiar_version}}{Version of the familiar package used to create this
object.}
}}

