% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Familiar.R
\name{get_xml_config}
\alias{get_xml_config}
\title{Create an empty xml configuration file}
\usage{
get_xml_config(dir_path)
}
\arguments{
\item{dir_path}{Path to the directory where the configuration file should be
created. The directory should exist, and no file named \code{config.xml} should
be present.}
}
\value{
Nothing. A file named \code{config.xml} is created in the directory
indicated by \code{dir_path}.
}
\description{
This function creates an empty configuration xml file in the directory
specified by \code{dir_path}. This provides an alternative to the use of input
arguments for familiar.
}
\examples{
\dontrun{
# Creates a config.xml file in the working directory
get_xml_config(dir_path=getwd())
}
}
\keyword{IO}
