% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{featureInfoParameters-class}
\alias{featureInfoParameters-class}
\title{Feature information parameters object.}
\description{
A featureInfo object contains information for a single feature. Some
information, for example concerning clustering and transformation contains
various parameters that allow for applying the data transformation correctly.
These are stored in featureInfoParameters objects.
}
\details{
featureInfoParameters is normally a parent class for specific
classes, such as featureInfoParametersTransformation.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the feature, which by default is the column name of the
feature. Typically used to correctly assign the data.}

\item{\code{complete}}{Flags whether the parameters have been completely set.}

\item{\code{familiar_version}}{Version of the familiar package.}
}}

