% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.update_experimental_design_settings}
\alias{.update_experimental_design_settings}
\title{Internal function to check batch assignment to development and validation}
\usage{
.update_experimental_design_settings(
  section_table,
  data,
  settings,
  verbose = FALSE
)
}
\arguments{
\item{section_table}{data.table generated by the \code{extract_experimental_setup}
function. Contains information regarding the experiment.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{settings}{List of parameter settings for data set parsing and setting
up the experiment.}
}
\value{
A verified and updated list of parameter settings.
}
\description{
This function checks which batches in the data set are assigned to model
development and external validation. Several errors may be raised if there
are inconsistencies such as an overlapping assignment, name mismatches etc.
}
\keyword{internal}
