% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarModel.R
\name{coef}
\alias{coef}
\alias{coef,familiarModel-method}
\title{Extract model coefficients}
\usage{
coef(object, ...)

\S4method{coef}{familiarModel}(object, ...)
}
\arguments{
\item{object}{a familiarModel object}

\item{...}{additional arguments passed to \code{coef} methods for the underlying
model, when available.}
}
\value{
Coefficients extracted from the model in the familiarModel object, if
any.
}
\description{
Extract model coefficients
}
\details{
This method extends the \code{coef} S3 method. For some models \code{coef}
requires information that is trimmed from the model. In this case a copy of
the model coefficient is stored with the model, and returned.
}
